/* ------------------------------------------------------------------------------

S061223 - HW ref: A241123 (SPP Printer Emulator Shield)


Tested on Atmega328P (Arduino UNO R3 and R2) @ Arduino IDE 1.8.19


NOTE1: the ACK output hanshaking signal follows the "Compatibility Mode" defined in the IEEE Std 1284-2000
       (see par. 7.3 of "IEEE Standard Signaling Method for a Bidirectional Parallel Peripheral
       Interface for Personal Computers" https://kazus.ru/nuke/modules/Downloads/pub/148/0/IEEE%201284-2000.pdf).

NOTE2: The ERROR and PAPEREND output signals are currently always not active (not used), and the AUTOFD and INIT
       input signals are ignored.


---------------------------------------------------------------------------------


CHANGELOG:


S061223           First revision.


--------------------------------------------------------------------------------- */


// ------------------------------------------------------------------------------
//
// Hardware definitions for A241123 board (see schematic)
//
// ------------------------------------------------------------------------------

// Centronics port data lines
#define   D0            12
#define   D1            11 
#define   D2            10 
#define   D3            9
#define   D4            8  
#define   D5            7
#define   D6            6
#define   D7            5

// Centronics port control lines
#define   STROBE_       3                           // STROBE input control line (active low)
#define   AUTOFD_       4                           // AUTOFD input control line (active low)
#define   INIT_         2                           // INIT input control line (active low)
#define   ACK_          13                          // ACK output control line (active low)
#define   BUSY          A2                          // BUSY output control line (active high)
#define   PAPEREND      A3                          // PAPEREND output control line (active high)
#define   SELECT        A4                          // SELECT output control line (active high)
#define   ERROR_        A5                          // ERROR output control line (active low)

// Led and key
#define   ON_LINE_LED   A0                          // ON-LINE led output (turned on if high)
#define   ON_LINE_KEY_  A1                          // ON-LINE push button input (low if pressed)

// Definitions for fast GPIO access
#define   POUT_1_BUSY   PORTC = PORTC | B00000100   // Set BUSY to 1
#define   POUT_0_BUSY   PORTC = PORTC & B11111011   // Set BUSY to 0
#define   POUT_1_ACK_   PORTB = PORTB | B00100000   // Set ACK to 1
#define   POUT_0_ACK_   PORTB = PORTB & B11011111   // Set ACK to 0
#define   PIN_KEY_      (PINC & B00000010)          // Read ON_LINE_KEY_
#define   TX_RDY        (UCSR0A & (1 << UDRE0))

// Other defines
#define   SW_ID         "S061223"                   // Software identification string (SW ID)
#define   DELAY_CYC(x)  __builtin_avr_delay_cycles(x)

// ------------------------------------------------------------------------------
//
//  Libraries
//
// ------------------------------------------------------------------------------

#include <EEPROM.h>                                 // Needed for internal EEPROM R/W

// ------------------------------------------------------------------------------
//
//  Constants
//
// ------------------------------------------------------------------------------

const long          serialSpeed     = 500000;
const unsigned int  debounceDelay   = 30; // Debounce time (milliseconds)
const byte          speedIdxAddr    = 10; // Internal EEPROM address for serial speed index storage
const unsigned int  blinkPeriodSlow = 600;// Time period for slow blinking (milliseconds)
const unsigned int  blinkPeriodFast = 100;// Time period for fast blinking (milliseconds)                                          

// ------------------------------------------------------------------------------
//
//  Global variables
//
// ------------------------------------------------------------------------------

enum printerState {offline = 0, online = 1};
enum buttonState  {notpressed, pressed};
volatile bool     dataReadyFlg;           // Data ready flag
volatile byte     PORTB_dataIn, PORTD_dataIn;
byte              PORTB_buffer, PORTD_buffer, rawDataToSend;
printerState      currentState;           // Current emulated printer status (online/offline)
unsigned long     timestamp, timestampLed;
buttonState       prevButton;             // Previos iteration key status (pressed/notpressed)
bool              actionDoneFlg;          // Action done flag
byte              i, convTable[256];
unsigned int      blinkPeriod;            // Time period for blinking leds (milliseconds)

// ------------------------------------------------------------------------------

void setup() 
{
  // Disable interrupts
  noInterrupts();
  
  // GPIO initialization
  pinMode(BUSY, OUTPUT);
  digitalWrite(BUSY, LOW);                // Set BUSY output not active
  pinMode(D0, INPUT);                     // Set D0-D7 data inputs
  pinMode(D1, INPUT);
  pinMode(D2, INPUT);
  pinMode(D3, INPUT);
  pinMode(D4, INPUT);
  pinMode(D5, INPUT);
  pinMode(D6, INPUT);
  pinMode(D7, INPUT);
  pinMode(STROBE_, INPUT);                // Set STROBE input
  pinMode(AUTOFD_, INPUT);                // Set AUTOFD input
  pinMode(INIT_, INPUT);                  // Set INIT input
  pinMode(ACK_, OUTPUT);
  digitalWrite(ACK_, HIGH);               // Set ACK not active
  pinMode(PAPEREND, OUTPUT);
  digitalWrite(PAPEREND, LOW);            // Set PAPEREND output not active
  pinMode(SELECT, OUTPUT);
  digitalWrite(SELECT, HIGH);             // Set SELECT output active
  pinMode(ERROR_, OUTPUT);
  digitalWrite(ERROR_, HIGH);             // Set ERROR output not active
  pinMode(ON_LINE_KEY_, INPUT_PULLUP);    // Set ON-LINE push button input

  // Set INT1 IRQ linked to the falling edge of the STROBE signal
  EICRA = 1 << ISC11;                     // Set to sense 1->0 change on the INT1 pin (STROBE_)
  EIMSK = 1 << INT1;                      // Enable INT1 (STROBE_) interrupt

  // Enable interrupts
  dataReadyFlg = false;
  interrupts();

  // Other settings
  pinMode(ON_LINE_LED, OUTPUT);           // Set ON-LINE led output
  digitalWrite(ON_LINE_LED, LOW);         // Set ON-LINE led off
  currentState = online;                  // Set printer status online
  prevButton = notpressed;
  actionDoneFlg = true;                   // Ignore if the ON-LINE key is pressed during power on
                                          // (used for the SW ID printing function)
  
  // Read the serial speed index and set the serial port speed accordingly
  if (EEPROM.read(speedIdxAddr) > 5)      // Check if it is a valid value, otherwise set it to default speed
  {
    EEPROM.update(speedIdxAddr, 4);       // Not a valid value. Set it to 500000 baud (default speed)
  }
  Serial.begin(indexToBaud(EEPROM.read(speedIdxAddr)));

  // Create the data byte conversion table.
  // This is used as a fast conversion method to take into account the physical layout of the 
  // Centronics data byte (D7-D0)
  i = 0;
  do
  {
    convTable[i] = ((i & 0b00010000) >> 4) |  // Raw data byte: D5 D6 D7 D0 D1 D2 D3 D4
                   ((i & 0b00001000) >> 2) |  //                b7 b6 b5 b4 b3 b2 b1 b0
                   ((i & 0b00000100) >> 0) |
                   ((i & 0b00000010) << 2) |
                   ((i & 0b00000001) << 4) |
                   ((i & 0b10000000) >> 2) |
                   ((i & 0b01000000) >> 0) |
                   ((i & 0b00100000) << 2);
    i++;
  }
  while (i > 0);
  
  // Enter into the setting mode if requested
  if (!digitalRead(ON_LINE_KEY_))
  {
    Serial.println();                     // Print the SW ID and current serial speed on the serial port
    Serial.print(SW_ID);
    Serial.print(" - Serial speed: ");
    Serial.print(indexToBaud(EEPROM.read(speedIdxAddr)));
    Serial.println("bps");
    do
    {
      // Blink ON_LINE_LED and BUSY leds accordingly with the current serial speed
      if (EEPROM.read(speedIdxAddr) > 2)
      {
        blinkPeriod = blinkPeriodFast;
      }
      else
      {
        blinkPeriod = blinkPeriodSlow;
      }
      if ((millis() - timestampLed) > blinkPeriod)
      {
        switch (EEPROM.read(speedIdxAddr))
        {
          case 0:                         // 9600/230400
          case 3:
            digitalWrite(BUSY, !digitalRead(BUSY));
          break;

          case 1:                         // 57600/500000
          case 4:
            digitalWrite(ON_LINE_LED, !digitalRead(ON_LINE_LED));
          break;

          case 2:                         // 115200/1000000
          case 5:
            digitalWrite(ON_LINE_LED, !digitalRead(ON_LINE_LED));
            digitalWrite(BUSY, digitalRead(ON_LINE_LED));
          break;
    
        }
        timestampLed = millis();
      }

      // Check ON-LINE key
      if (!digitalRead(ON_LINE_KEY_))
      // ON-LINE key pressed, so debounce it and change the current printer status if valid
      {
        if (prevButton == notpressed)
        // Key was pressed first time, so set the timestamp for debounce
        {
          timestamp = millis();
          prevButton = pressed;
        }
        if (((millis() - timestamp) > debounceDelay) && !actionDoneFlg)
        // End of debounce delay, so it is a valid action
        {
          EEPROM.update(speedIdxAddr, ((EEPROM.read(speedIdxAddr) +1) % 6));
          digitalWrite(ON_LINE_LED, LOW);
          digitalWrite(BUSY, LOW);
          timestampLed = millis();
          actionDoneFlg = true;           // Prevent continuous toggling if button remains pressed
        }
      }
      else
      {
        prevButton = notpressed;
        actionDoneFlg = false;
      }
    }
    while (1);
  }
  digitalWrite(ON_LINE_LED, HIGH);        // Set ON-LINE led on
  TIMSK0 = 0;                             // Disable the Timer0 interrupt to avoid to interfere with the INT1 latency time.
                                          // Now all the time related funcions are disabled.
  UCSR0B = UCSR0B & ~(1 << UDRIE0);       // Disable the UDR0 register empty interrupt to avoid to interfere with the INT1 latency time.
                                          // Now all the Serial TX related funcions are disabled.                
}

// ------------------------------------------------------------------------------

void loop()
{
  // Do parallel to serial print
  if (dataReadyFlg && (TX_RDY>0) && (currentState == online))
  // There is a byte ready to send/print, the serial Tx buffer isn't full and the printer status is on-line, so
  // send it to the serial port
  {
    
    // Save data to send and reset the data ready flag
    PORTB_buffer = PORTB_dataIn;
    PORTD_buffer = PORTD_dataIn;
    dataReadyFlg = false;

    // ACK/BUSY handshaking sequence (Centronics port)
    noInterrupts();                       // Just to be sure. Not really needed as only INT1 is enabled and not triggered here
    POUT_0_ACK_;                          // Set ACK active
    DELAY_CYC(16);                        // Delay 1us (16 * 62.5ns)
    POUT_1_ACK_;                          // Set ACK not active
    interrupts();
    POUT_0_BUSY;                          // Set BUSY not active. Now a new data byte can be received from the Centronics port

    // Send data to serial after reordering bit (see schematic)
    rawDataToSend = PORTB_buffer & 0b00011111;                    // Raw data byte:  0  0  0 D0 D1 D2 D3 D4
    rawDataToSend = rawDataToSend | (PORTD_buffer & 0B11100000);  // Raw data byte: D5 D6 D7 D0 D1 D2 D3 D4
    UDR0 = convTable[rawDataToSend];                              //                b7 b6 b5 b4 b3 b2 b1 b0
  }

  // Check ON-LINE key
  if (!PIN_KEY_)
  // ON-LINE key pressed, so debounce it and change the current printer status if valid
  {
    if (!PIN_KEY_ && !actionDoneFlg)
    {
      DELAY_CYC((unsigned long) ((debounceDelay * 1000000) / 62.5)); // Debounce delay
      if (!PIN_KEY_ && !actionDoneFlg)
      // End of debounce delay, so it is a valid action
      {
        currentState = (printerState) !currentState;  // Toggle current "online" status
        if (currentState == offline)
        // Set the BUSY signal and ON_LINE_LED accordingly
        {
          digitalWrite(BUSY, HIGH);
          digitalWrite(ON_LINE_LED, LOW);
        }
        else 
        {
          if (!dataReadyFlg)
          // If there is a previous pending char to be printed (dataReadyFlg = true) the
          // BUSY signal remains not active to avoid a new char to overwrite it. The pending char will
          // be printed during the next loop (if the output buffer is not full) and then the BUSY signal 
          // set not active
          {
            digitalWrite(BUSY, LOW);
          }
          digitalWrite(ON_LINE_LED, HIGH);
        }
        actionDoneFlg = true;             // Prevent continuous toggling if button remains pressed
      }
    }
  }
  else
  {
    actionDoneFlg = false;
  }
}

// ------------------------------------------------------------------------------

ISR(INT1_vect, ISR_NAKED)
// Interrupt service routine for INT1 (first part using assembler).
// To minimize the interrupt latency time a low level approach is used here.
// For more info see here: https://arduino.stackexchange.com/questions/8758/arduino-interruption-on-pin-change
{
  asm volatile(
  "    sbi %[gpio], 2         \n"         // sbi  PORTC, 2            ; Set BUSY active (high)
  "    rjmp INT1_vect_part_2  \n"         // rjmp INT1_vect_part_2    ; Jump to second part
  :: [gpio] "I" (_SFR_IO_ADDR(PORTC)));
}

ISR(INT1_vect_part_2)
// Interrupt service routine for INT1 (second part)
{
  // Read data from Centronics port
  PORTB_dataIn = PINB;
  PORTD_dataIn = PIND;
  
  // Set data ready flag
  dataReadyFlg = true;
}

// ------------------------------------------------------------------------------

unsigned long indexToBaud(byte index)
// Output the real baud rate for a given speed index value.
// Valid values for the speed index are in the [0..5] range
{
  unsigned long   baudRate;
  switch (index)
  {
    case 0: 
      baudRate = 9600;
    break;

    case 1: 
      baudRate = 57600;
    break;
    
    case 2: 
      baudRate = 115200;
    break;
    
    case 3: 
      baudRate = 230400;
    break;
    
    case 4: 
      baudRate = 500000;
    break;

    default:
      baudRate = 1000000;
    break;
  }
  return baudRate;
}
